module sequence_input 
  !
  ! Purpopse: 
  !   Collect datatypes and procedures for reading sequence data from disk.
  ! Copyright (C) 2004 Bernhard Haubold
  ! Software distributed under the terms of the GNU GENERAL PUBLIC LICENSE
  ! Record of revisions:
  !    Date            Programmer           Description of change
  !    ====            ==========           =====================
  !  May 02, 2004      B. Haubold           Original code
  !
  !
  implicit none 
  ! Define basic type for representing a set of strings
  type text_row
     character(1), pointer, dimension(:) :: r         ! One row of text
  end type text_row
  contains
subroutine read_fasta(u, text, names, max_seq, max_line)
  !
  ! Purpopse: 
  !   Read sequence data from fasta file. In its current form the maximum
  !   length a sequence may have is 5 million nucleoties. This needs to
  !   be revised.
  ! Record of revisions:
  !    Date            Programmer           Description of change
  !    ====            ==========           =====================
  !  Jan. 05, 2004     B. Haubold           Original code
  !
  implicit none
  integer   , intent(in)                :: u              ! The unit to read from
  type(text_row), pointer, dimension(:) :: text           ! The concatenated sequence data
  character(256), pointer, dimension(:) :: names          ! The names of the sequences
  integer, intent(in)                   :: max_seq        ! The maximum number of sequences accepted
  integer, intent(in)                   :: max_line       ! The maximum line length accepted
  character(max_line+1)                 :: string
  integer    ios, len, c, i, total_len
  integer, allocatable, dimension(:)    :: lengths          ! Lengths of the sequences
  integer, parameter                    :: scratch_unit=99  ! Unit for scratch file
  !
  ! Code
  !
  allocate(lengths(max_seq))
  lengths(:)=0
  ! Count the number of sequences
  c=0
  len=0
  ! Open scatch file for buffering data
  ! Find out number of sequences in file and their lengths
  do
     ! Check for end of file
     read(u,iostat=ios,fmt='(a)')string
     if(ios.eq.-1)exit
     if(scan(string,'>').eq.1) then
        c=c+1
        if(c.gt.max_seq) then
           write(6,'(a,i12)') 'ERROR in read_fasta: maximum number of sequences is ', max_seq
           stop
        end if
     else
        if(len_trim(string).gt.max_line) then
           write(6,'(a,i12)') 'ERROR in read_fasta: maximum length of lines in FASTA file is ', max_line
           stop
        end if
        lengths(c) = lengths(c) + len_trim(string)
     end if
  end do
  rewind(u)
  ! Allocate memory
  allocate(text(c))
  allocate(names(c))
  do i=1,c
     allocate(text(i)%r(lengths(i)))
  end do
  c=1
  len=0
  ! Read first definition line
  read(u,iostat=ios,fmt='(a)') string
  names(c)=string(2:len_trim(string))
  do
     ! Read one line from file
     read(u,iostat=ios,fmt='(a)') string
     ! Check for header line
     if(scan(string,'>').eq.0.and.ios.eq.0) then
        do i=1,len_trim(string)
           text(c)%r(len+i) = string(i:i)
        end do
        len = len + len_trim(string)
     elseif(ios.eq.0)then
        len=0
        c=c+1
        names(c)=string(2:len_trim(string))
     else
        exit
     end if
  end do
  deallocate(lengths)
end subroutine read_fasta

subroutine prepare_data(text, molecule_type, sentinel)
  !
  ! Purpopse: 
  !    Prepare the input data for suffix tree construction by doing the following three things:
  !       1) Eliminate case by setting all characters to upper case.
  !       2) Add sentinel character to the end of each sequence, unless the sentinel 
  !          is the empty character ''.
  !       3) Consider the "molecule_type". This may take the following values:
  !          1 = nucleotide, linear, double stranded (default)
  !          2 = nucleotide, linear, single stranded
  !          3 = nucleotide, circular double stranded
  !          4 = nucleotide, circular single stranded
  !          5 = other
  !    1 = nucleotide, double stranded,    - upper case, filter, & add rev.-comp.
  !    2 = nucleotide, single stranded,    - upper case, filter
  !    3 = other                           - upper case
  ! Record of revisions:
  !    Date            Programmer           Description of change
  !    ====            ==========           =====================
  !  Jan. 25, 2004     B. Haubold           Original code
  !
  use string_utils
  implicit none
  type(text_row), pointer, dimension(:)   :: text           ! The sequence data
  integer, intent(in)                     :: molecule_type  ! Type of molecule passed
  character(1), intent(in)                :: sentinel       ! Character to be added at end of sequence
  integer                                 :: i, isize, rsize, i2, left, right
  type(text_row), pointer                 :: row
  character(1), pointer, dimension(:)     :: char_array
  type(text_row), pointer, dimension(:)   :: text2          ! Construct new text here
  integer, parameter                      :: bridge_len = 50! Length of bridging sequence; use EVEN number

  ! Put sequences to upper case and set strand
  isize = size(text)
  do i=1, isize
     call tra_uc(text(i)%r)
  end do
  ! Filter nucleotides?
  if(molecule_type.lt.5)then
     do i=1, isize
        call char_sar('R','G',text(i)%r)
        call char_sar('Y','T',text(i)%r)
        call char_sar('M','A',text(i)%r)
        call char_sar('K','G',text(i)%r)
        call char_sar('S','G',text(i)%r)
        call char_sar('W','A',text(i)%r)
        call char_sar('H','A',text(i)%r)
        call char_sar('B','G',text(i)%r)
        call char_sar('V','G',text(i)%r)
        call char_sar('D','G',text(i)%r)
        call char_sar('N','G',text(i)%r)
        call char_sar('U','T',text(i)%r)
     end do
  end if
  ! Double-stranded molecule? In this case add reverse-complement to data set
  if(molecule_type.eq.1.or.molecule_type.eq.3) then
     allocate(text2(isize*2))
     do i=1,isize
        i2=(i-1)*2+1
        text2(i2)%r=>text(i)%r
        allocate(char_array(size(text(i)%r)))
        char_array(:)=text2(i2)%r(:)
        call reverse_complement(char_array)
        text2(i2+1)%r=>char_array
     end do
     text=>text2
  end if
  ! Add sentinel? In this case the length of each row needs to be extended by one
  if(sentinel.ne.'')then
     isize = size(text)
     do i=1, isize
        rsize = size(text(i)%r)
        allocate(char_array(rsize+1))
        char_array(1:rsize)=text(i)%r(:)
        deallocate(text(i)%r)
        text(i)%r=>char_array
        text(i)%r(rsize+1)=sentinel
     end do
  end if

end subroutine prepare_data

end module sequence_input
