module stree
  implicit none
  !
  ! Purpopse: 
  !   Define data types and subroutines for constructing generalized suffix trees.
  ! Copyright (C) 2004 Bernhard Haubold
  ! Software distributed under the terms of the GNU GENERAL PUBLIC LICENSE
  ! Record of revisions:
  !    Date            Programmer           Description of change
  !    ====            ==========           =====================
  !  May 02, 2004      B. Haubold           Original code
  !  Dec 08, 2005      B. Haubold           add_child: Add child nodes in alphabetical order
  !
  !
  ! Define an extensible leaf label to provide for construction of generalized strees
  type leaf_label
     type(leaf_label), pointer :: next      ! The next leaf_label
     integer                   :: id        ! Identifies the string of origin
     integer                   :: pos       ! Start position of suffix in string of origin
  end type leaf_label
  !
  ! Define basic tree type
  type node
     type(node), pointer                 :: parent       ! References parent node
     type(node), pointer                 :: child        ! References first child node
     type(node), pointer                 :: sibling      ! References sibling
     type(node), pointer                 :: link         ! Suffix link
     type(leaf_label), pointer           :: leaf         ! Leaf label (only used if node is leaf)
     character(1), pointer, dimension(:) :: label        ! The edge label
     integer                  :: depth           ! Number of characters on the path from root to this node; negative
     !   value used to identify source node.
     integer                  :: id              ! Node identifier; used for debugging purposes (small int sufficient)
  end type node
  !
  ! Define node for drawing tree; same as node except for x-position and y-position
  type draw_node
     type(draw_node), pointer                 :: parent       ! References parent node
     type(draw_node), pointer                 :: child        ! References first child node
     type(draw_node), pointer                 :: sibling      ! References sibling
     type(draw_node), pointer                 :: link         ! Suffix link
     type(leaf_label), pointer                :: leaf         ! Leaf label (only used if node is leaf)
     real                                     :: x, y         ! X- & y-positions of node
     character(1), pointer, dimension(:)      :: label        ! The edge label
     integer                  :: depth        ! Number of characters on the path from root to this node; negative
     !   value used to identify source node.
     integer                  :: id           ! Node identifier; used for debugging purposes (small int sufficient)
  end type draw_node
  !
  ! Define the reference pair for the suffix
  ! currently being inserted into the tree
  type reference_pair
     type(node), pointer                 :: node      ! Node referring to the node BELOW which the sought suffix ends
     character(1), dimension(:), pointer :: suffix    ! The suffix
     logical                             :: end_point ! Indicates whether or not reference pair refers to end point
  end type reference_pair
  interface generate_node
     module procedure generate_node1, generate_node2, generate_node3
  end interface
contains
  subroutine construct_stree(text,root)
    !
    ! Purpopse: 
    !   Construction of a suffix tree.
    !   Reference: E. Ukkonen (1995). On-line construction of suffix trees. Algorithmica
    !              14, 249-260.
    ! Record of revisions:
    !    Date            Programmer           Description of change
    !    ====            ==========           =====================
    !  Feb. 01, 2004     B. Haubold           Original code
    !
    use sequence_input
    implicit none
    type(text_row), dimension(:)             :: text     ! The text corresponding to the suffix tree
    type(node), pointer                      :: root     ! The root of the suffix tree
    type(node), pointer                      :: source
    character(1), pointer, dimension(:)      :: row
    integer    i, text_size


    ! Generate root node
    call generate_node(source)
    call generate_node(root)
    source%child=>root
    ! A negative depth value is used to identify the source node
    source%depth=-1
    ! Root has node depth of zero
    root%depth=0
    ! Set the label to a zero-length text for source
    source%label=>text(1)%r(1:0)
    ! Set the label to unit-length text for root
    root%label=>text(1)%r(-1:-1)
    ! Create suffix link between root and source
    root%link=>source
    ! Construct suffix tree
    text_size=size(text)
    do i=1,text_size
       row=>text(i)%r
       call add_string(row, i, root)
    enddo
  end subroutine construct_stree

  subroutine add_string(text, id, root)
    !
    ! Purpopse: 
    !   Insert new string "text" identified by "id" into suffix tree accessible via its root node.
    !   Reference: E. Ukkonen (1995). On-line construction of suffix trees. Algorithmica
    !              14, 249-260.
    ! Record of revisions:
    !    Date            Programmer           Description of change
    !    ====            ==========           =====================
    !  Jan. 05, 2004     B. Haubold           Original code
    !  Feb. 09, 2004     B. Haubold           Fixed inclusion of duplicate strings
    !
    implicit none
    character(1), pointer, dimension(:)                 :: text     ! The text corresponding to the suffix tree
    integer   , intent(in)                              :: id       ! Text identifier
    type(node), pointer                                 :: root     ! The root of the suffix tree
    type(node), pointer                                 :: child
    type(reference_pair), pointer                       :: rp       ! Reference pair
    integer                                             :: right_border
    type(leaf_label), pointer                           :: ll
    type(node), pointer                                 :: test_node
    character(1), pointer                               :: t_i
    integer    i, k, n, leaf, suffix_len

    ! Generate reference pair
    allocate(rp)
    ! Let the reference pair refer to the tree's root
    rp%node=>root
    k = 1
    n = size(text)
    leaf = 1
    call find_right_border(root, text, rp, right_border)
    ! The following if-block takes care of insertion of duplicate strings
    if(right_border.gt.n)then
       if(.not.associated(rp%node%child))then
          i=1
          child=>rp%node
          rp%node=>rp%node%parent
          call update_leaves(rp, child, i, id)
       end if
    end if
    do i=right_border, n
       suffix_len = size(rp%suffix)
       call update(rp, root, text, i, id, leaf)
       ! Readjust k
       suffix_len = size(rp%suffix)
       call canonize(rp)
       rp%suffix=>rp%suffix(1:size(rp%suffix)+1)
    end do
  end subroutine add_string

  subroutine find_right_border(root, text, rp, rb)
    !
    ! Purpopse: 
    !    Find the "index" of the first position in "text" that
    !    has a mismatch in the suffix tree represented by "root".
    ! Record of revisions:
    !    Date            Programmer           Description of change
    !    ====            ==========           =====================
    !  Jan. 07, 2004     B. Haubold           Original code
    !
    implicit none
    type(node), pointer                                 :: root        ! The root of the suffix tree
    character(1), pointer, dimension(:)                 :: text        ! The text corresponding to the suffix tree
    type(reference_pair), pointer                       :: rp          ! The reference pair
    integer   , intent(out)                             :: rb          ! The index of the first mismatch (right border)
    integer    i
    type(node), pointer                                 :: child
    character(1), pointer                               :: t

    rp%node=>root
    t=>text(1)
    ! Set the right border
    rb = 1   
    call find_child(rp%node, child, t)
    if(.not.associated(child)) then
       rp%suffix=>text(1:1)
       return
    else
       rp%node=>child
       rp%suffix=>text(1:1)
    end if
    ! Set index into edge labels
    i = 1
    node_hopping: do
       ! If entire text was found in tree or if there is a mismatch
       if(rb.gt.size(text).or.rp%node%label(i).ne.text(rb)) then
          ! If the last match ended at the end of an edge label
          if(i-1.eq.size(rp%node%label)) then
             ! Reduce reference suffix to its last character and exit
             rp%suffix=>rp%suffix(size(rp%suffix):)
             exit node_hopping
          else
             ! Return parent of current reference node
             rp%node=>rp%node%parent
             exit node_hopping
          end if
       end if
       i = i+1
       rb = rb+1
       rp%suffix=>rp%suffix(1:size(rp%suffix)+1)
       ! Has the edge-label index grown beyond the size of the edge label?
       if(i.gt.size(rp%node%label)) then
          t=>text(rb)
          ! Find the next node away from the root
          call find_child(rp%node, child, t)
          ! Is the text pointer (rb for right border) still inside the text? and
          ! Has a child been found?
          if(rb.lt.size(text).and.associated(child)) then
             ! Adjust suffix
             rp%suffix=>rp%suffix(size(rp%node%label)+1:)
             ! Move the reference node to the child node
             rp%node=>child
             ! Reset the edge index
             i=1
             ! If no child has been found, the match ended at the end of an edge. 
             ! In this case, reduce referece suffix to its last character and exit loop.
          else
             rp%suffix=>rp%suffix(size(rp%suffix):)
             exit node_hopping
          end if
       end if
    end do node_hopping
  end subroutine find_right_border

  subroutine update_leaves(rp, child, leaf, id)
    !
    ! Purpopse: 
    !   This subroutine is specific to generalized suffix trees. It inserts additional
    !   leaf labels into existing leaves at which suffices of a new string end.
    ! Record of revisions:
    !    Date            Programmer           Description of change
    !    ====            ==========           =====================
    !  Jan. 09, 2004     B. Haubold           Original code
    !
    implicit none
    type(reference_pair), pointer        :: rp    ! Reference pair
    type(node), pointer                  :: child ! t_k child-node of the current reference node
    integer   , intent(inout)            :: leaf  ! Leaf label
    integer   , intent(in)               :: id    ! String identifier
    type(leaf_label), pointer            :: ll
    type(reference_pair), pointer        :: new_rp ! Copy of the reference pair

    ! Generate new leaf label...
    allocate(ll)
    ! ...indicating string of origin...
    ll%id = id
    ! ...and starting position of the suffix.
    ll%pos = leaf
    leaf = leaf+1
    ! Add the leaf label to the leaf
    call add_leaf_label(child, ll)
    ! Search for other leaves to update
    allocate(new_rp)
    new_rp%suffix=>child%label
    new_rp%node=>rp%node
    do
       if(size(new_rp%suffix).eq.0) exit
       new_rp%node=>new_rp%node%link
       call canonize(new_rp)
       ! Generate a leaf label...
       allocate(ll)
       ! ...indicating string of origin...
       ll%id = id
       ! ...and starting position of the suffix.
       ll%pos = leaf
       leaf = leaf+1
       ! Add the leaf label to the leaf
       call add_leaf_label(new_rp%node, ll)
       ! Unless we are dealing with "root", obtain the label of the canonized node...
       if(associated(new_rp%node%parent)) then
          new_rp%suffix=>new_rp%node%label
          ! ...and hop to the canonized node's parent
          new_rp%node=>new_rp%node%parent
       end if
    end do
  end subroutine update_leaves

  subroutine update(rp, root, text, i, id, leaf)
    !
    ! Purpopse: 
    !   Updating a suffix tree.
    !   Reference: E. Ukkonen (1995). On-line construction of suffix trees. Algorithmica
    !              14, 249-260.
    ! Record of revisions:
    !    Date            Programmer           Description of change
    !    ====            ==========           =====================
    !  Jan. 06, 2004     B. Haubold           Original code
    !
    implicit none
    type(reference_pair), pointer        :: rp   ! Reference pair
    type(node), pointer                  :: root ! Root of stree
    character(1), pointer, dimension(:)  :: text ! Text of suffix tree
    integer   , intent(in)             :: i    ! Index of character to be inserted
    integer   , intent(in)             :: id   ! String identifier
    integer   , intent(inout)          :: leaf ! Keep track of leaf label
    character(1), pointer                :: t_i  ! Character to be inserted
    type(node), pointer                  :: oldr ! Old root
    type(node), pointer                  :: s, r_dash
    type(leaf_label), pointer            :: ll
    type(node), pointer                  :: test_node
    character(1), pointer                :: test_char

    t_i=>rp%suffix(size(rp%suffix))
    ! Canonize reference pair
    rp%suffix=>rp%suffix(1:size(rp%suffix)-1)
    oldr=>root
    ! Keep reference to incoming node
    s=>rp%node
    call test_and_split(rp, t_i, leaf, id)
    do
       if(rp%end_point) exit
       ! Generate new leaf
       call generate_node(r_dash, rp%node, text(i:))
       test_char=>text(i)
       ! Generate a leaf label...
       allocate(ll)
       ! ...indicating string of origin...
       ll%id = id
       ! ...and starting position of the suffix.
       ll%pos = leaf
       nullify(ll%next)
       leaf = leaf+1
       ! Add the leaf label to the leaf
       call add_leaf_label(r_dash, ll)
       ! If the current active node is not root, create suffix link
       if(associated(oldr%parent)) oldr%link=>rp%node
       oldr=>rp%node
       ! Follow suffix link
       rp%node=>s%link
       call canonize(rp)
       s=>rp%node
       call test_and_split(rp, t_i, leaf, id)
    end do
    ! Fix suffix link
    if(associated(oldr%parent)) then
       oldr%link=>s
    endif
    ! Expand label again
    rp%suffix=>rp%suffix(1:size(rp%suffix)+1)
  end subroutine update

  subroutine test_and_split(rp, t, leaf, id)
    !
    ! Purpopse: 
    !   The purpose of this routine is twofold:
    !   1) It makes the current reference point explicit, if it is not already so.
    !   2) It tests whether the current reference point refers to the end-point, i.e.
    !      is followed by an implicit or explicit state referring to the character 
    !      currently being inserted.
    !   Reference: E. Ukkonen (1995). On-line construction of suffix trees. Algorithmica
    !      14, 249-260.
    ! Record of revisions:
    !    Date            Programmer           Description of change
    !    ====            ==========           =====================
    !  Jan. 05, 2004     B. Haubold           Original code
    !
    implicit none
    type(reference_pair), pointer     :: rp    ! The reference pair
    character(1), pointer             :: t     ! The last letter of the suffix to be inserted
    integer                         :: leaf  ! Leaf label
    integer                         :: id    ! String identifier
    type(node), pointer               :: child
    type(node), pointer               :: r
    type(leaf_label), pointer         :: ll
    character(1), pointer             :: t_k
    type(reference_pair), pointer     :: new_rp

    ! Check if the reference pairs' suffix has length greater than 0
    if(size(rp%suffix).gt.0) then
       t_k => rp%suffix(1)
       ! Find the t_k transition from the reference node
       call find_child(rp%node, child, t_k)
       ! Test whether the edge label of "child" contains character t, the character to be inserted
       if(.not.associated(child)) write(6,*) 'rp%node%id ',rp%node%id,' is not associated via ',t_k, &
            ' StringId: ', id
       if(child%label(1+size(rp%suffix)).eq.t) then
          ! If so, the end point has been found
          rp%end_point = .true.
          ! Take care of generalized suffix tree in the following if block
          ! by checking whether the match ended at the last position of an 
          ! edge label ending at a leaf
          if(.not.associated(child%child).and.1+size(rp%suffix).eq.size(child%label)) then
             call update_leaves(rp, child, leaf, id)
          end if
          return
       else
          ! Otherwise, the reference pair referring to character t needs to be made explicit:
          ! Split edge,...
          call extract_child(rp%node, child, t_k)
          call generate_node(r, rp%node, child, child%label(1:size(rp%suffix)), child%label(size(rp%suffix)+1:))
          ! ...reset the node of the reference pair, and specify that the end point is NOT reached
          rp%node=>r
          rp%end_point=.false.
          ! Specify node depth
          r%depth = r%parent%depth + size(r%label)
          return
       end if
    else ! State already explicit, so no splitting necessary
       call find_child(rp%node, child, t)
       if(.not.associated(child)) then
          rp%end_point=.false.
          return
       else
          rp%end_point=.true.
          ! Take care of generalized suffix tree in the following if block
          if(.not.associated(child%child).and.size(child%label).eq.1) then
             call update_leaves(rp, child, leaf, id)
          end if
          return
       end if
    end if
  end subroutine test_and_split


  subroutine canonize(rp)
    !
    ! Purpopse: 
    !   Find the position of a suffix in the suffix tree.
    ! Record of revisions:
    !    Date            Programmer           Description of change
    !    ====            ==========           =====================
    !  Jan. 05, 2004     B. Haubold           Original code
    !
    implicit none
    type(reference_pair), pointer     :: rp    ! The reference pair
    type(node), pointer               :: child
    character(1), pointer             :: t_k

    ! Check if reference pair is already canonized
    if(size(rp%suffix).eq.0) then
       return
    endif
    ! Otherwise, perform canonization
    t_k =>  rp%suffix(1)
    call find_child(rp%node, child, t_k)
    do
       if(size(child%label).gt.size(rp%suffix)) then
          exit
       end if
       ! Cut off prefix the size of child's label from the reference pair label
       rp%suffix=>rp%suffix(1+size(child%label):)
       ! Update node of reference pair
       rp%node=>child
       if(size(rp%suffix).gt.0) then
          t_k =>  rp%suffix(1)
          call find_child(rp%node, child, t_k)
       endif
    end do
  end subroutine canonize

  subroutine find_child(parent, child, c)
    !
    ! Purpopse: 
    !   Find a child node identified by the first character
    !   of its edge label.
    ! Record of revisions:
    !    Date            Programmer           Description of change
    !    ====            ==========           =====================
    !  Jan. 05, 2004     B. Haubold           Original code
    !
    implicit none
    type(node), pointer               :: parent
    type(node), pointer               :: child
    character(1), pointer             :: c
    type(node), pointer               :: current

    if(.not.associated(parent%child)) then
       nullify(child)
       return
    else if(parent%depth.lt.0) then
       ! Parent is source node, in which case the first child, which is the root,
       ! is always returned
       child=>parent%child
    else
       ! Deal with first child
       current=>parent%child
       if(current%label(1).eq.c) then
          child=>current
          return
       else
          ! Deal with remaining children
          do
             if(associated(current%sibling)) then
                current=>current%sibling
             else
                nullify(child)
                return
             endif
             if(current%label(1).eq.c) then
                child=>current
                return
             end if
          end do
       end if
    end if
  end subroutine find_child

  subroutine generate_node1(n)
    !
    ! Purpopse: 
    !   Generate a new node with all links nullified.
    ! Record of revisions:
    !    Date            Programmer           Description of change
    !    ====            ==========           =====================
    !  Jan. 05, 2004     B. Haubold           Original code
    !  Jan. 23, 2004     B. Haubold           Added initial nullification of leaf label 
    !
    implicit none
    type(node), pointer      :: n
    integer                  :: i

    allocate(n)
    call label(i)
    n%id = i
    n%depth=0
    nullify(n%parent)
    nullify(n%child)
    nullify(n%sibling)
    nullify(n%link)
    nullify(n%leaf)
  end subroutine generate_node1

  subroutine generate_node2(n, parent, node_label)
    !
    ! Purpopse: 
    !   Generate a new node with parent node specified
    ! Record of revisions:
    !    Date            Programmer           Description of change
    !    ====            ==========           =====================
    !  Jan. 05, 2004     B. Haubold           Original code
    !
    implicit none
    type(node), pointer                      :: n, parent
    character(1), dimension(:), target       :: node_label
    integer                                  :: i

    allocate(n)
    call label(i)
    n%id = i
    n%depth=0
    n%parent=>parent
    n%label=>node_label;
    nullify(n%sibling)
    call add_child(parent, n)
    nullify(n%child)
    nullify(n%link)
    nullify(n%leaf)
  end subroutine generate_node2

  subroutine generate_node3(n, parent, child, node_label, child_label)
    !
    ! Purpopse: 
    !   Generate a new node with parent and first child nodes specified
    ! Record of revisions:
    !    Date            Programmer           Description of change
    !    ====            ==========           =====================
    !  Jan. 05, 2004     B. Haubold           Original code
    !
    implicit none
    type(node), pointer                      :: n, parent, child
    character(1), dimension(:), target       :: node_label, child_label
    integer                                  :: i

    allocate(n)
    n%label=>node_label
    child%label=>child_label
    call label(i)
    n%id = i
    n%depth=0
    n%parent=>parent
    nullify(n%child)
    nullify(n%sibling)
    call add_child(parent, n)
    call add_child(n, child)
    nullify(n%link)
    nullify(n%leaf)
  end subroutine generate_node3

  subroutine label(i)
    !
    ! Purpopse: 
    !   Generate an integer label that grows on each invocation of the subroutine
    ! Record of revisions:
    !    Date            Programmer           Description of change
    !    ====            ==========           =====================
    !  Jan. 06, 2004     B. Haubold           Original code
    !
    implicit none
    integer   , intent(inout) :: i
    integer                   :: x = 0
    x = x + 1
    i = x
  end subroutine label
  subroutine add_child(parent, child)
    !
    ! Purpopse: 
    !   Add child node to suffix tree.
    ! Record of revisions:
    !    Date            Programmer           Description of change
    !    ====            ==========           =====================
    !  Jan. 05, 2004     B. Haubold           Original code
    !  Dec. 08, 2005     B. Haubold           Add children in alphabetical order
    !
    implicit none
    type(node), pointer               :: parent
    type(node), pointer               :: child
    type(node), pointer               :: n

    ! Prepare child for insertion
    child%parent=>parent
    nullify(child%sibling)
    ! Search for insertion position
    if(.not.associated(parent%child)) then
       parent%child=>child
    else
       if(ichar(child%label(1)).lt.ichar(parent%child%label(1)))then
          child%sibling=>parent%child
          parent%child=>child
       else
          n=>parent%child
          do
             if(.not.associated(n%sibling))then
                n%sibling=>child
                exit
             else if(ichar(child%label(1)).lt.ichar(n%sibling%label(1)))then
                child%sibling=>n%sibling
                n%sibling=>child
                exit
             else
                n=>n%sibling
             end if
          end do
       end if
       end if
  end subroutine add_child
  subroutine add_leaf_label(leaf, ll)
    !
    ! Purpopse: 
    !   Add a new element to the list of leaf labels.
    ! Record of revisions:
    !    Date            Programmer           Description of change
    !    ====            ==========           =====================
    !  Jan. 05, 2004     B. Haubold           Original code
    !
    implicit none
    type(node), pointer               :: leaf
    type(leaf_label), pointer         :: ll
    type(leaf_label), pointer         :: l      ! Dummy leaf label

    ! Test whether the incoming leaf has any leaf labels at all
    if(.not.associated(leaf%leaf)) then
       leaf%leaf=>ll
    else   
       ! Associate dummy with current leaf label
       l=>leaf%leaf
       ! Move dummy along list until end is found
       do
          if(associated(l%next)) then
             l=>l%next
          else
             l%next=>ll
             nullify(ll%next)
             exit
          end if
       end do
    end if
  end subroutine add_leaf_label
  subroutine extract_child(parent, child, c)
    !
    ! Purpopse: 
    !   Extract a child node identified by the first character
    !   of its edge label. Extraction means that the child's pointers
    !   to "parent" and "sibling" are nullified, but not its pointer
    !   to "child" (if assiciated).
    ! Record of revisions:
    !    Date            Programmer           Description of change
    !    ====            ==========           =====================
    !  Jan. 06, 2004     B. Haubold           Original code
    !
    implicit none
    type(node), pointer               :: parent
    type(node), pointer               :: child
    character(1), pointer             :: c
    type(node), pointer               :: previous

    if(.not.associated(parent%child)) then
       stop 'ERROR in extract_child: dealing with leaf node.'
    else
       ! Test first child
       child=>parent%child
       previous=>parent%child
       if(child%label(1).eq.c) then
          ! Dissociate the child from the parent
          if(associated(child%sibling)) then
             parent%child=>child%sibling
          else
             nullify(parent%child)
          end if
       else
          ! Test remaining children
          do
             if(associated(child%sibling)) then
                ! Move along list of children
                previous=>child
                child=>child%sibling
             else
                exit
             endif
             if(child%label(1).eq.c) then
                ! Distinguish between position within...
                if(associated(child%sibling)) then
                   previous%sibling=>child%sibling
                else
                   ! ...and at end of list
                   nullify(previous%sibling)
                end if
                exit
             end if
          end do
       end if
       ! Dissociate the parent from the child
       nullify(child%parent)
       ! Dissociate the child's siblings
       nullify(child%sibling)
    end if
  end subroutine extract_child

  recursive subroutine destruct_stree(root)
    !
    ! Purpopse: 
    !   Free the memory taken up by a suffix tree rooted on "root"
    ! Record of revisions:
    !    Date            Programmer           Description of change
    !    ====            ==========           =====================
    !  Feb. 01, 2004     B. Haubold           Original code
    !
    implicit none
    type(node), pointer                      :: root     ! The root of the suffix tree
    type(node), pointer                      :: child, sibling

    child=>root%child
    do
       if(.not.associated(child))exit
       sibling=>child%sibling
       call destruct_stree(child)
       child=>sibling
    end do
    ! Remove leaf node
    if(associated(root%parent)) root%parent%child=>root%sibling
    deallocate(root)
  end subroutine destruct_stree

end module stree
